@@ -5,6 +5,7 @@
 
 #include "config.h"
 
+#include "xkbcommon/xkbcommon.h"
 #include "xkbcomp-priv.h"
 #include "text.h"
 #include "vmod.h"
@@ -297,14 +298,14 @@ AddMapEntry(KeyTypesInfo *info, KeyTypeInfo *type,
         if (report && old->level != new->level) {
             log_warn(info->ctx, XKB_WARNING_CONFLICTING_KEY_TYPE_MAP_ENTRY,
                      "Multiple map entries for %s in %s; "
-                     "Using %d, ignoring %d\n",
+                     "Using %"PRIu32", ignoring %"PRIu32"\n",
                      MapEntryTxt(info, new), TypeTxt(info, type),
                      (clobber ? new->level : old->level) + 1,
                      (clobber ? old->level : new->level) + 1);
         }
         else {
             log_vrb(info->ctx, 10, XKB_WARNING_CONFLICTING_KEY_TYPE_MAP_ENTRY,
-                    "Multiple occurrences of map[%s]= %d in %s; Ignored\n",
+                    "Multiple occurrences of map[%s]= %"PRIu32" in %s; Ignored\n",
                     MapEntryTxt(info, new), new->level + 1,
                     TypeTxt(info, type));
             return true;
@@ -487,7 +488,7 @@ AddLevelName(KeyTypesInfo *info, KeyTypeInfo *type,
     /* Same level, same name. */
     if (darray_item(type->level_names, level) == name) {
         log_vrb(info->ctx, 10, XKB_WARNING_DUPLICATE_ENTRY,
-                "Duplicate names for level %d of key type %s; Ignored\n",
+                "Duplicate names for level %"PRIu32" of key type %s; Ignored\n",
                 level + 1, TypeTxt(info, type));
         return true;
     }
@@ -499,7 +500,7 @@ AddLevelName(KeyTypesInfo *info, KeyTypeInfo *type,
                             darray_item(type->level_names, level));
         new = xkb_atom_text(info->ctx, name);
         log_vrb(info->ctx, 1, XKB_WARNING_CONFLICTING_KEY_TYPE_LEVEL_NAMES,
-                "Multiple names for level %d of key type %s; "
+                "Multiple names for level %"PRIu32" of key type %s; "
                 "Using %s, ignoring %s\n",
                 level + 1, TypeTxt(info, type),
                 (clobber ? new : old), (clobber ? old : new));
@@ -508,7 +509,7 @@ AddLevelName(KeyTypesInfo *info, KeyTypeInfo *type,
             return true;
     }
 
-    /* XXX: What about different level, same name? */
+    /* FIXME: What about different level, same name? */
 
 finish:
     darray_item(type->level_names, level) = name;
@@ -530,7 +531,7 @@ SetLevelName(KeyTypesInfo *info, KeyTypeInfo *type, ExprDef *arrayNdx,
     xkb_atom_t level_name = XKB_ATOM_NONE;
     if (!ExprResolveString(info->ctx, value, &level_name)) {
         log_err(info->ctx, XKB_ERROR_WRONG_FIELD_TYPE,
-                "Non-string name for level %d in key type %s; "
+                "Non-string name for level %"PRIu32" in key type %s; "
                 "Ignoring illegal level name definition\n",
                 level + 1, xkb_atom_text(info->ctx, type->name));
         return false;
@@ -635,6 +636,35 @@ HandleKeyTypeDef(KeyTypesInfo *info, KeyTypeDef *def)
     return true;
 }
 
+static bool
+HandleGlobalVar(KeyTypesInfo *info, VarDef *stmt)
+{
+    const char *elem, *field;
+    ExprDef *arrayNdx;
+
+    if (!ExprResolveLhs(info->ctx, stmt->name, &elem, &field,
+                        &arrayNdx))
+        return false;           /* internal error, already reported */
+
+    if (elem && istreq(elem, "type")) {
+        log_err(info->ctx, XKB_ERROR_WRONG_STATEMENT_TYPE,
+                "Support for changing the default type has been removed; "
+                "Statement ignored\n");
+        return true;
+    }
+    else if (elem) {
+        log_err(info->ctx, XKB_ERROR_UNKNOWN_DEFAULT_FIELD,
+                "Default defined for unknown element \"%s\"; "
+                "Value for field \"%s.%s\" ignored\n", elem, elem, field);
+    }
+    else if (field) {
+        log_err(info->ctx, XKB_ERROR_UNKNOWN_DEFAULT_FIELD,
+                "Default defined for unknown field \"%s\"; Ignored\n", field);
+    }
+
+    return false;
+}
+
 static void
 HandleKeyTypesFile(KeyTypesInfo *info, XkbFile *file)
 {
@@ -652,10 +682,7 @@ HandleKeyTypesFile(KeyTypesInfo *info, XkbFile *file)
             ok = HandleKeyTypeDef(info, (KeyTypeDef *) stmt);
             break;
         case STMT_VAR:
-            log_err(info->ctx, XKB_ERROR_WRONG_STATEMENT_TYPE,
-                    "Support for changing the default type has been removed; "
-                    "Statement ignored\n");
-            ok = true;
+            ok = HandleGlobalVar(info, (VarDef*) stmt);
             break;
         case STMT_VMOD:
             ok = HandleVModDef(info->ctx, &info->mods, (VModDef *) stmt);
@@ -685,11 +712,10 @@ HandleKeyTypesFile(KeyTypesInfo *info, XkbFile *file)
 static bool
 CopyKeyTypesToKeymap(struct xkb_keymap *keymap, KeyTypesInfo *info)
 {
-    unsigned num_types;
-    struct xkb_key_type *types;
-
-    num_types = darray_empty(info->types) ? 1 : darray_size(info->types);
-    types = calloc(num_types, sizeof(*types));
+    const darray_size_t num_types = darray_empty(info->types)
+                                  ? 1
+                                  : darray_size(info->types);
+    struct xkb_key_type *types = calloc(num_types, sizeof(*types));
     if (!types)
         return false;
 
@@ -709,14 +735,16 @@ CopyKeyTypesToKeymap(struct xkb_keymap *keymap, KeyTypesInfo *info)
         type->num_level_names = 0;
     }
     else {
-        for (unsigned i = 0; i < num_types; i++) {
+        for (darray_size_t i = 0; i < num_types; i++) {
             KeyTypeInfo *def = &darray_item(info->types, i);
             struct xkb_key_type *type = &types[i];
 
             type->name = def->name;
             type->mods.mods = def->mods;
             type->num_levels = def->num_levels;
-            darray_steal(def->level_names, &type->level_names, &type->num_level_names);
+            type->num_level_names =
+                (xkb_level_index_t) darray_size(def->level_names);
+            darray_steal(def->level_names, &type->level_names, NULL);
             darray_steal(def->entries, &type->entries, &type->num_entries);
         }
     }
