@@ -11,14 +11,85 @@
 #include <stdio.h>
 #include <stdlib.h>
 
+#include "xkbcommon/xkbcommon.h"
+#include "xkbcommon/xkbcommon-keysyms.h"
 #include "evdev-scancodes.h"
 #include "test.h"
-#include "xkbcommon/xkbcommon-keysyms.h"
 #include "keymap.h"
+#include "keymap-formats.h"
+#include "utils.h"
 
 #define KEY_LVL3 84
 #define KEY_LVL5 195
 
+static void
+test_supported_formats(void)
+{
+    assert(xkb_keymap_parse_format(NULL) == 0);
+    assert(xkb_keymap_parse_format("") == 0);
+    assert(xkb_keymap_parse_format("x") == 0);
+    assert(xkb_keymap_parse_format("v") == 0);
+    assert(xkb_keymap_parse_format("vx") == 0);
+    assert(xkb_keymap_parse_format("0x1") == 0); /* only base 10 */
+    assert(xkb_keymap_parse_format("+1") == XKB_KEYMAP_FORMAT_TEXT_V1);
+    assert(xkb_keymap_parse_format(" 1") == XKB_KEYMAP_FORMAT_TEXT_V1);
+
+    const struct {
+        int value;
+        enum xkb_keymap_format expected;
+        const char* const* labels;
+    } entries[] = {
+        { .value = -1                            , .labels = NULL, .expected = 0 },
+        { .value = 0                             , .labels = NULL, .expected = 0 },
+        { .value = 100000000                     , .labels = NULL, .expected = 0 },
+        { .value = XKB_KEYMAP_USE_ORIGINAL_FORMAT, .labels = NULL, .expected = 0 },
+        {
+          .value = XKB_KEYMAP_FORMAT_TEXT_V1,
+          .labels = (const char* const[]) { "v1", NULL },
+          .expected = XKB_KEYMAP_FORMAT_TEXT_V1
+        },
+        {
+          .value = XKB_KEYMAP_FORMAT_TEXT_V2,
+          .labels = (const char* const[]) { "v2", NULL },
+          .expected = XKB_KEYMAP_FORMAT_TEXT_V2
+        },
+    };
+    char buf[15] = { 0 };
+    for (size_t k = 0; k < ARRAY_SIZE(entries); k++) {
+        assert(xkb_keymap_is_supported_format(entries[k].value) ==
+               !!entries[k].expected);
+        /* Parse labels */
+        for (size_t l = 0; entries[k].labels && entries[k].labels[l]; l++) {
+            assert_printf(xkb_keymap_parse_format(entries[k].labels[l]) ==
+                          entries[k].expected,
+                          "%s: expected %d, got: %d\n",
+                          entries[k].labels[l], entries[k].value,
+                          xkb_keymap_parse_format(entries[k].labels[l]));
+        }
+        /* Parse serialized numeric value */
+        const int ret = snprintf(buf, sizeof(buf), "%d", entries[k].value);
+        assert(ret > 0 && ret < (int) sizeof(buf));
+        assert(xkb_keymap_parse_format(buf) == entries[k].expected);
+    }
+
+    const enum xkb_keymap_format *formats;
+    const size_t count = xkb_keymap_supported_formats(&formats);
+    assert(count == 2);
+    enum xkb_keymap_format previous = 0; /* Lower bound */
+    for (size_t k = 0; k < count; k++) {
+        /* Ascending order */
+        assert(previous < formats[k]);
+        /* Valid numeric format */
+        assert(xkb_keymap_is_supported_format(formats[k]));
+        /* Valid label */
+        const int ret = snprintf(buf, sizeof(buf), "%d", formats[k]);
+        assert(ret > 0 && ret < (int) sizeof(buf));
+        assert(xkb_keymap_parse_format(buf) == formats[k]);
+
+        previous = formats[k];
+    }
+}
+
 static void
 test_garbage_key(void)
 {
@@ -32,7 +103,8 @@ test_garbage_key(void)
 
     assert(context);
 
-    keymap = test_compile_rules(context, NULL, NULL, "garbage", NULL, NULL);
+    keymap = test_compile_rules(context, XKB_KEYMAP_FORMAT_TEXT_V1, NULL,
+                                NULL, "garbage", NULL, NULL);
     assert(keymap);
 
     /* TLDE uses the 'us' sym on the first level and is thus [grave, exclam] */
@@ -77,7 +149,8 @@ test_keymap(void)
 
     assert(context);
 
-    keymap = test_compile_rules(context, "evdev", "pc104", "us,ru", NULL, "grp:menu_toggle");
+    keymap = test_compile_rules(context, XKB_KEYMAP_FORMAT_TEXT_V1, "evdev",
+                                "pc104", "us,ru", NULL, "grp:menu_toggle");
     assert(keymap);
 
     kc = xkb_keymap_key_by_name(keymap, "AE09");
@@ -147,8 +220,8 @@ test_no_extra_groups(void)
     assert(context);
 
     /* RMLVO: Legacy rules may add more layouts than the input RMLVO */
-    keymap = test_compile_rules(context, "multiple-groups",
-                                "old", "de", NULL, NULL);
+    keymap = test_compile_rules(context, XKB_KEYMAP_FORMAT_TEXT_V1,
+                                "multiple-groups", "old", "de", NULL, NULL);
     assert(keymap);
     kc = xkb_keymap_key_by_name(keymap, "AD01");
     assert(kc != XKB_KEYCODE_INVALID);
@@ -161,8 +234,9 @@ test_no_extra_groups(void)
     const char *layouts[] = {"us", "us,us", "us,us,us", "us,us,us,us"};
     for (xkb_layout_index_t k = 0; k < ARRAY_SIZE(layouts); k++) {
         /* `multiple-groups` option defines 4 groups for a key */
-        keymap = test_compile_rules(context, "multiple-groups", NULL,
-                                    layouts[k], NULL, "multiple-groups");
+        keymap = test_compile_rules(context, XKB_KEYMAP_FORMAT_TEXT_V1,
+                                    "multiple-groups", NULL, layouts[k], NULL,
+                                    "multiple-groups");
         assert(keymap);
         kc = xkb_keymap_key_by_name(keymap, "RALT");
         assert(kc != XKB_KEYCODE_INVALID);
@@ -180,8 +254,8 @@ test_no_extra_groups(void)
 
     /* RMLVO: Ensure the rule “one group per key” in symbols sections works
      * for the 2nd layout */
-    keymap = test_compile_rules(context, NULL, NULL,
-                                "multiple-groups,multiple-groups", "1,2", NULL);
+    keymap = test_compile_rules(context, XKB_KEYMAP_FORMAT_TEXT_V1, NULL,
+                                NULL, "multiple-groups,multiple-groups", "1,2", NULL);
     assert(keymap);
     kc = xkb_keymap_key_by_name(keymap, "RALT");
     assert(kc != XKB_KEYCODE_INVALID);
@@ -206,7 +280,7 @@ test_no_extra_groups(void)
         "  xkb_symbols { include \"pc+multiple-groups(1)+multiple-groups(2):2"
                                   "+inet(evdev)\" };"
         "};";
-    keymap = test_compile_string(context, keymap_str);
+    keymap = test_compile_string(context, XKB_KEYMAP_FORMAT_TEXT_V1, keymap_str);
     kc = xkb_keymap_key_by_name(keymap, "RALT");
     assert(kc != XKB_KEYCODE_INVALID);
     assert(xkb_keymap_num_layouts_for_key(keymap, kc) == 4);
@@ -239,7 +313,8 @@ test_numeric_keysyms(void)
 
     assert(context);
 
-    keymap = test_compile_rules(context, "evdev", "pc104", "numeric_keysyms", NULL, NULL);
+    keymap = test_compile_rules(context, XKB_KEYMAP_FORMAT_TEXT_V1, "evdev",
+                                "pc104", "numeric_keysyms", NULL, NULL);
     assert(keymap);
 
     kc = xkb_keymap_key_by_name(keymap, "AD01");
@@ -282,7 +357,8 @@ test_multiple_keysyms_per_level(void)
 
     assert(context);
 
-    keymap = test_compile_rules(context, "evdev", "pc104", "awesome", NULL, NULL);
+    keymap = test_compile_rules(context, XKB_KEYMAP_FORMAT_TEXT_V1, "evdev",
+                                "pc104", "awesome", NULL, NULL);
     assert(keymap);
 
     kc = xkb_keymap_key_by_name(keymap, "AD01");
@@ -330,8 +406,8 @@ test_multiple_actions_per_level(void)
     assert(context);
 
     /* Test various ways to set multiple actions */
-    keymap = test_compile_rules(context, "evdev", "pc104",
-                                "multiple_actions,cz", NULL, NULL);
+    keymap = test_compile_rules(context, XKB_KEYMAP_FORMAT_TEXT_V1, "evdev",
+                                "pc104", "multiple_actions,cz", NULL, NULL);
     assert(keymap);
 
     kc = xkb_keymap_key_by_name(keymap, "LCTL");
@@ -468,8 +544,8 @@ test_multiple_actions_per_level(void)
      *       layout. However, this requires to be able to configure group redirect
      *       at the *keymap* level, then use ISO_First_Group and SetGroup(group=-4).
      *       Change the symbols and this test once this is merged. */
-    keymap = test_compile_rules(context, "evdev", "pc104",
-                                "awesome,cz", NULL, "grp:menu_toggle");
+    keymap = test_compile_rules(context, XKB_KEYMAP_FORMAT_TEXT_V1, "evdev",
+                                "pc104", "awesome,cz", NULL, "grp:menu_toggle");
     assert(keymap);
 
     ctrl = xkb_keymap_mod_get_index(keymap, XKB_MOD_NAME_CTRL);
@@ -531,6 +607,7 @@ main(void)
 {
     test_init();
 
+    test_supported_formats();
     test_garbage_key();
     test_keymap();
     test_no_extra_groups();
